// Copyright (C) 2022 ASTRON (Netherlands Institute for Radio Astronomy)
// SPDX-License-Identifier: GPL-3.0-or-later

#define BOOST_TEST_MODULE schaapcommon_fitters

#include <boost/test/unit_test.hpp>