// Copyright (C) 2023 ASTRON (Netherlands Institute for Radio Astronomy)
// SPDX-License-Identifier: GPL-3.0-or-later

#include "meerkatcoefficients.h"

namespace everybeam {
namespace circularsymmetric {
constexpr std::array<double, 6> MeerKATCoefficients::coefficients_;
}
}  // namespace everybeam
