// Copyright (C) 2021 ASTRON (Netherlands Institute for Radio Astronomy)
// SPDX-License-Identifier: GPL-3.0-or-later

#include "atcacoefficients.h"

namespace everybeam {
namespace circularsymmetric {
constexpr std::array<double, 35> ATCACoefficients::coefficients_;
}
}  // namespace everybeam
